# Copyright (c) 2024 Snowflake Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from pathlib import Path
from typing import List, Literal, Optional, Union

from pydantic import Field
from snowflake.cli.api.project.schemas.entities.common import (
    EntityBase,
)
from snowflake.cli.api.project.schemas.native_app.package import DistributionOptions
from snowflake.cli.api.project.schemas.native_app.path_mapping import PathMapping
from snowflake.cli.api.project.schemas.updatable_model import IdentifierField


class ApplicationPackageEntity(EntityBase):
    type: Literal["application package"]  # noqa: A003
    name: str = Field(
        title="Name of the application package created when this entity is deployed"
    )
    artifacts: List[Union[PathMapping, Path]] = Field(
        title="List of paths or file source/destination pairs to add to the deploy root",
    )
    bundle_root: Optional[Path] = Field(
        title="Folder at the root of your project where artifacts necessary to perform the bundle step are stored.",
        default=Path("output/bundle/"),
    )
    deploy_root: Optional[Path] = Field(
        title="Folder at the root of your project where the build step copies the artifacts",
        default=Path("output/deploy/"),
    )
    generated_root: Optional[Path] = Field(
        title="Subdirectory of the deploy root where files generated by the Snowflake CLI will be written.",
        default=Path("__generated/"),
    )
    stage: Optional[str] = IdentifierField(
        title="Identifier of the stage that stores the application artifacts.",
        default="app_src.stage",
    )
    scratch_stage: Optional[str] = IdentifierField(
        title="Identifier of the stage that stores temporary scratch data used by the Snowflake CLI.",
        default="app_src.stage_snowflake_cli_scratch",
    )
    distribution: Optional[DistributionOptions] = Field(
        title="Distribution of the application package created by the Snowflake CLI",
        default="internal",
    )
    manifest: Path = Field(
        title="Path to manifest.yml",
    )
